#include <time.h>
#include <string.h>
#include <unistd.h>

#include <list>

#include "clients.h"
#include "md5.h"

ClientDB::ClientDB()
{
}

ClientDB::~ClientDB()
{
}

bool ClientDB::load(int fd)
{
	int i,n;
	int size;
	time_t limit;
	Client *p;
	std::string id;
	
	limit = time(NULL) - 86400;

	n = lseek(fd, 0, SEEK_END) / sizeof(Client);
	lseek(fd, 0, SEEK_SET);

	size = n*sizeof(Client);
	p = (Client *)malloc(size);
	if(p == NULL){
		return false;
	}

	read(fd, p, size);

	for(i=0;i<n;i++){
		if( (p[i].last_access < limit) &&
		    (p[i].work_count < 1) ){
			// ȏo߂ĕ񍐐 1 ȉ
			// ꍇ͓o^
			continue;
		}
		id.assign(((char *)(p[i].id)), 16);
		this->db.insert(std::make_pair(id,p[i]));
	}

	free(p);

	return true;
}

void ClientDB::save(int fd)
{
	int n;
	int idx;
	long size;
	Client *p;
	
	std::map<std::string, Client>::iterator i;

	n = this->db.size();
	size = n*sizeof(Client);
	p = (Client *)malloc(size);

	idx = 0;
	for(i=this->db.begin();i!=this->db.end();i++){
		memcpy(p+idx, &(i->second), sizeof(Client));
		idx += 1;
	}

	lseek(fd, 0, SEEK_SET);
	write(fd, p, size);
	ftruncate(fd, size);

	return;
}

std::string ClientDB::newId()
{
	time_t now;
	struct tm *tp;
	char buffer[128];
	
	std::string id;
	Client elem;

	MD5 md5;

	time(&now);

	do {
		now += 1;
		tp = localtime(&now);
		strftime(buffer, sizeof(buffer), "%a %b %d %H:%M:%S %Z %Y", tp);

		md5.init();
		md5.update((const int8_t *)buffer, strlen(buffer));
		id = md5.digest();

	} while( this->db.find(id) != this->db.end() );

	memcpy(elem.id, id.c_str(), 16);
	elem.last_access = time(NULL);
	elem.work_count = 0;

	this->db.insert(std::make_pair(id, elem));

	return id;
}

Client *ClientDB::get(std::string &id)
{
	std::map<std::string, Client>::iterator pos;

	pos = this->db.find(id);
	if(pos != this->db.end()){
		return &(pos->second);
	}

	return NULL;
}

int ClientDB::countActiveClient()
{
	int r;
	time_t limit;
	std::map<std::string, Client>::iterator i;

	r = 0;
	limit = time(NULL)-21600;
	for(i=this->db.begin();i!=this->db.end();i++){
		if(i->second.last_access >= limit){
			r += 1;
		}
	}

	return r;
}

int ClientDB::countAllClient()
{
	return this->db.size();
}

static inline std::string clientInfo(Client &c, time_t limit)
{
	int i;
	int active;
	char id[33];
	char msg[128];

	for(i=0;i<16;i++){
		sprintf(id+i*2, "%02x", c.id[i]);
	}

	if( limit < c.last_access ){
		active = 1;
	}else{
		active = 0;
	}

	sprintf(msg, "%8d:%s [%d]\n", c.work_count, id, active);

	return msg;
}

void ClientDB::printAllClient()
{
	time_t limit;
	std::map<std::string, Client>::iterator i;
	std::list<std::string>::iterator j;
	std::list<std::string> info;

	limit = time(NULL)-86400;
	for(i=this->db.begin();i!=this->db.end();i++){
		info.push_back(clientInfo(i->second, limit));
	}
	info.sort();
	for(j=info.begin();j!=info.end();j++){
		fprintf(stdout, "%s", j->c_str());
	}
}

int ClientDB::rank(std::string &id)
{
	int r;
	Client *elem;
	std::map<std::string, Client>::iterator i;

	r = 1;
	elem = this->get(id);
	if(elem == NULL){
		return -1;
	}
	for(i=this->db.begin();i!=this->db.end();i++){
		if(elem != &(i->second)){
			if(elem->work_count < i->second.work_count){
				r += 1;
			}
		}
	}

	return r;
}
