#include <time.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <gd.h>
#include <gdfonts.h>

#include "status.h"
#include "c2.h"
#include "table.h"
#include "log.h"

static void setup_work(Work &unit);
static bool check_work(Work &master, Work &report);

static void dump(Work &w, char *msg);
static bool debug_check(Work &w, int idx);

Status::Status()
{
}

Status::~Status()
{
}

bool Status::load(int fd)
{
	int n;
	int sr,sw;
	int32_t i;
	int32_t work_count;
	int32_t complete_count;

	std::list<Range>::iterator back,it;

	uint32_t limit;
	
	Range *r;
	Work  *w;
	
	this->work.clear();
	this->complete.clear();

	lseek(fd, 0, SEEK_SET);

	n = read(fd, &complete_count, sizeof(int32_t));
	if(n != sizeof(int32_t)){
		return false;
	}
	   
	n = read(fd, &work_count, sizeof(int32_t));
	if(n != sizeof(int32_t)){
		return false;
	}

	sr = complete_count * sizeof(Range);
	r = (Range *)malloc(sr);
	if(r == NULL){
		return false;
	}

	sw = work_count * sizeof(Work);
	w = (Work *)malloc(sw);
	if(w == NULL){
		free(r);
		return false;
	}
			
	n = read(fd, r, sr);
	if(n != sr){
		free(r);
		free(w);
		return false;
	}

	n = read(fd, w, sw);
	if(n != sw){
		free(r);
		free(w);
		return false;
	}
	
	for(i=0;i<complete_count;i++){
		this->complete.push_back(r[i]);
	}
	
	limit = time(NULL) - 86400;
	for(i=0;i<work_count;i++){
		if(w[i].time >= limit){
			// 蓖Ă 1 ȓL
			this->work.insert(std::make_pair(w[i].block, w[i]));
		}
	}

	this->complete.sort();

	// marge list
	it = this->complete.begin();
	back = it;
	it++;
	while(it != this->complete.end()){
		if(back->tail >= it->head){
			it->head = back->head;
			if(it->tail < back->tail){
				it->tail = back->tail;
			}
			this->complete.erase(back);
		}
		back = it;
		it++;
	}

	free(r);
	free(w);

	return true;
}

void Status::save(int fd)
{
	int n;
	int sr,sw;
	int idx;
	int32_t work_count;
	int32_t complete_count;

	Range *r;
	Work  *w;

	std::list<Range>::iterator i;
	std::map<uint32_t, Work>::iterator j;
	
	complete_count = this->complete.size();
	work_count = this->work.size();
	
	sr = complete_count * sizeof(Range);
	r = (Range *)malloc(sr);
	if(r == NULL){
		return;
	}

	sw = work_count * sizeof(Work);
	w = (Work *)malloc(sw);
	if(w == NULL){
		free(r);
		return;
	}

	idx = 0;
	for(i=this->complete.begin();i!=this->complete.end();i++){
		r[idx] = *i;
		idx += 1;
	}

	idx = 0;
	for(j=this->work.begin();j!=this->work.end();j++){
		memcpy(w+idx, &(j->second), sizeof(Work));
		idx += 1;
	}

	lseek(fd, 0, SEEK_SET);

	write(fd, &complete_count, sizeof(int32_t));
	write(fd, &work_count, sizeof(int32_t));
	write(fd, r, sr);
	write(fd, w, sw);
	ftruncate(fd, sr+sw+sizeof(int32_t)+sizeof(int32_t));
}

std::list<Work> Status::getBlock(std::string &id, int count)
{
	uint32_t key;
	uint32_t now;
	Work w;
	
	std::list<Work> r;
	std::list<Range>::iterator pos;

	pos = this->complete.begin();
	key = 0;
	now = time(NULL);
	
	while(count > 0){
		if( key >= 256*256*256 ){
			break;
		}
		if( pos != this->complete.end() ){
			if(key >= pos->head){
				if(key < pos->tail){
					key = pos->tail;
				}
				pos++;
				continue;
			}
		}
		if( this->work.find(key) == this->work.end() ){
			memcpy(w.id, id.c_str(), sizeof(w.id));
			w.block = key;
			w.time = now;
			setup_work(w);
			this->work.insert(std::make_pair(w.block,w));
			r.push_back(w);
			count -= 1;
		}
		key += 1;
	}

	return r;
}

int Status::report(std::string &id, Work &unit)
{
	std::map<uint32_t, Work>::iterator w;
	
	Range r;

	w = this->work.find(unit.block);
	if(w == this->work.end()){
		logout("Status::report() - no block=%06x\n", unit.block);
		return 0;
	}
	if(memcmp(w->second.id, id.c_str(), sizeof(w->second.id)) != 0){
		dump(w->second, "Status::report() - id missmatch");
		return 0;
	}
	if(!check_work(w->second, unit)){
		dump(w->second, "Status::report() - failed on check_work()");
		this->work.erase(w);
		return -1;
	}
	this->work.erase(w);

	r.head = unit.block;
	r.tail = unit.block+1;

	this->complete.push_back(r);

	return 1;
}

bool Status::cancel(uint32_t block)
{
	Range w1,w2;
	std::list<Range>::iterator it;
	
	it = this->complete.begin();
	
	while(it != this->complete.end()){
		if(it->head == block){
			w1.head = it->head + 1;
			w1.tail = it->tail;
			this->complete.erase(it);
			if(w1.head != w1.tail){
				this->complete.push_back(w1);
			}
			return true;
		}else if( (it->head < block) && (block < it->tail) ){
			w1.head = it->head;
			w1.tail = block;
			w2.head = block+1;
			w2.tail = it->tail;
			this->complete.erase(it);
			this->complete.push_back(w1);
			this->complete.push_back(w2);
			return true;
		}
		it++;
	}

	return false;
}


void Status::updateImage(int workingClient)
{
	FILE *png;
	gdImagePtr im;

	int black,white;
	
	int i;
	int cb;
	int head,tail;

	struct stat st;

	char buf[256];

	std::list<Range>::iterator pos;

	stat("current.png", &st);
	if(st.st_mtime > time(NULL)-60){
		// ŏIXV 60 bȓ͍XVXLbv
		return;
	}

	im = gdImageCreate(256,256);
	if(im == NULL){
		return;
	}

	white = gdImageColorAllocate(im, 255, 255, 255);
	black = gdImageColorAllocate(im, 0, 0, 0);

	cb = 0;
	pos = this->complete.begin();
	while(pos != this->complete.end()){
		head = pos->head;
		tail = pos->tail;
		cb += tail-head;
		head >>= 8;
		tail += 255;
		tail >>= 8;
		for(i=head;i<tail;i++){
			gdImageSetPixel(im, i&255, i>>8, black);
		}
		pos++;
	}

	if( cb < (128*256*256) ){
		sprintf(buf, "working clients:  %8d", workingClient);
		gdImageString(im, gdFontSmall, 5, 205, (unsigned char *)buf, black);
		sprintf(buf, "working blocks:   %8d", this->work.size());
		gdImageString(im, gdFontSmall, 5, 220, (unsigned char *)buf, black);
		sprintf(buf, "completed blocks: %8d (%.02f%%)", cb, cb*100.0/(256*256*256));
		gdImageString(im, gdFontSmall, 5, 235, (unsigned char *)buf, black);
	}else{
		sprintf(buf, "working clients:  %8d", workingClient);
		gdImageString(im, gdFontSmall, 5,  5, (unsigned char *)buf, white);
		sprintf(buf, "working blocks:   %8d", this->work.size());
		gdImageString(im, gdFontSmall, 5, 20, (unsigned char *)buf, white);
		sprintf(buf, "completed blocks: %8d (%.02f%%)", cb, cb*100.0/(256*256*256));
		gdImageString(im, gdFontSmall, 5, 35, (unsigned char *)buf, white);
	}

	png = fopen("current.png", "wb");
	if(png == NULL){
		gdImageDestroy(im);
		return;
	}

	gdImagePng(im, png);

	fclose(png);
	gdImageDestroy(im);
}

void Status::debug()
{
	int idx;
	std::list<Range>::iterator i;
	std::map<uint32_t, Work>::iterator j;

	char id[33];
	
	for(i=this->complete.begin();i!=this->complete.end();i++){
		fprintf(stdout, "c: head=%06x, tail=%06x, count=%d\n", i->head, i->tail, i->tail-i->head);
	}

	for(j=this->work.begin();j!=this->work.end();j++){
		for(idx=0;idx<16;idx++){
			sprintf(id+idx*2, "%02x", j->second.id[idx]);
		}
		fprintf(stdout, "w: id=%s, key=%06x, time=%d\n", id, j->second.block, j->second.time);
	}
}

static void setup_work(Work &unit)
{
	int i,j,n;
	uint64_t pt;
	uint32_t wk;
	uint32_t kb[8];
	uint32_t dk[8];

	pt = strtoull(PLAIN_TEXT, NULL, 16);
	
	n = (rand() >> 13) & 0x7;
	if(n > 7){
		n = 7;
	}else if(n < 2){
		n = 2;
	}

	for(i=0;i<n;i++){
		wk = ((uint32_t)rand()) << 16;
		wk |= rand();
		kb[i] = wk;
		dk[i] = 1;
	}

	for(i=0;i<n-1;i++){
		for(j=i+1;j<n;j++){
			if(kb[i] == kb[j]){
				dk[j] = 0;
			}
		}
	}

	for(i=0,j=0;i<n;i++){
		if(dk[i]){
			unit.key[j] = (((uint64_t)unit.block)<<32) | kb[i];
			unit.code[j] = c2_enc(pt, unit.key[j]);
			j += 1;
		}
	}

	unit.count = j;

	for(i=j;i<7;i++){
		wk = (((uint32_t)rand()) << 16) | rand();
		unit.key[i] = (((uint64_t)unit.block+i)<<32) | wk;
		unit.code[i] = c2_enc(pt, unit.key[i]);
	}
	
	unit.code[7] = strtoull(CYPHER_TEXT, NULL, 16);
}

static bool check_work(Work &master, Work &report)
{
	int i,j;
	bool ff[8];

	for(i=0;i<master.count;i++){
		ff[i] = false;
	}

	for(i=0;i<master.count;i++){
		for(j=0;j<report.count;j++){
			if(master.key[i] == report.key[j]){
				ff[i] = true;
			}
		}
	}

	for(i=0;i<master.count;i++){
		if(ff[i] == false){
			return false;
		}
	}

	return true;
}

static void dump(Work &w, char *msg)
{
	int i;
	
	logout("%s\n", msg);
	logout("ID: ");
	for(i=0;i<16;i++){
		logout("%02x", w.id[i]);
	}
	logout("\n");
		
	logout("count=%d\n", w.count);
	logout("block=%06x\n", w.block);
	logout("time=%08x\n", w.time);
	for(i=0;i<8;i++){
		logout("key=%014llx, code=%016llx\n", w.key[i], w.code[i]);
	}
}

static bool debug_check(Work &w, int idx)
{
	int i;
	uint64_t text;
	uint64_t code;

	char msg[1024];
	
	if(w.count > 7){
		sprintf(msg, "[%d] Work::count overflow", idx);
		dump(w, msg);
		return false;
	}

	if(w.block > 0xffffff){
		sprintf(msg, "[%d] Work::block overflow", idx);
		dump(w, msg);
		return false;
	}

	text = strtoull(PLAIN_TEXT, NULL, 16);

	for(i=0;i<w.count;i++){
		if( (w.key[i] >> 32) != w.block ){
			sprintf(msg, "[%d] Work::key[%d] missmatch Work::block[%d]", idx, i, i);
			dump(w, msg);
			return false;
		}
		code = c2_enc(text, w.key[i]);
		if(code != w.code[i]){
			sprintf(msg, "[%d] Work::code[%d] missmatch Work::key[%d]", idx, i, i);
			dump(w, msg);
			return false;
		}
	}

	return true;
}


